from enum import StrEnum
from typing import Optional, Self

from .config import VeilidConfig
from .types import (
    ByteCount,
    RouteId,
    Timestamp,
    TimestampDuration,
    TypedKey,
    ValueData,
    ValueSubkey,
    VeilidLogLevel,
    OperationId,
    urlsafe_b64decode_no_pad,
)


class AttachmentState(StrEnum):
    DETACHED = "Detached"
    ATTACHING = "Attaching"
    ATTACHED_WEAK = "AttachedWeak"
    ATTACHED_GOOD = "AttachedGood"
    ATTACHED_STRONG = "AttachedStrong"
    FULLY_ATTACHED = "FullyAttached"
    OVER_ATTACHED = "OverAttached"
    DETACHING = "Detaching"


class VeilidStateAttachment:
    state: AttachmentState
    public_internet_ready: bool
    local_network_ready: bool

    def __init__(
        self,
        state: AttachmentState,
        public_internet_ready: bool,
        local_network_ready: bool,
    ):
        self.state = state
        self.public_internet_ready = public_internet_ready
        self.local_network_ready = local_network_ready

    @classmethod
    def from_json(cls, j: dict) -> Self:
        """JSON object hook"""
        return cls(
            AttachmentState(j["state"]),
            j["public_internet_ready"],
            j["local_network_ready"],
        )


class RPCStats:
    messages_sent: int
    messages_rcvd: int
    questions_in_flight: int
    last_question_ts: Optional[Timestamp]
    last_seen_ts: Optional[Timestamp]
    first_consecutive_seen_ts: Optional[Timestamp]
    recent_lost_answers: int
    failed_to_send: int

    def __init__(
        self,
        messages_sent: int,
        messages_rcvd: int,
        questions_in_flight: int,
        last_question_ts: Optional[Timestamp],
        last_seen_ts: Optional[Timestamp],
        first_consecutive_seen_ts: Optional[Timestamp],
        recent_lost_answers: int,
        failed_to_send: int,
    ):
        self.messages_sent = messages_sent
        self.messages_rcvd = messages_rcvd
        self.questions_in_flight = questions_in_flight
        self.last_question_ts = last_question_ts
        self.last_seen_ts = last_seen_ts
        self.first_consecutive_seen_ts = first_consecutive_seen_ts
        self.recent_lost_answers = recent_lost_answers
        self.failed_to_send = failed_to_send

    @classmethod
    def from_json(cls, j: dict) -> Self:
        """JSON object hook"""
        return cls(
            j["messages_sent"],
            j["messages_rcvd"],
            j["questions_in_flight"],
            None if j["last_question_ts"] is None else Timestamp(j["last_question_ts"]),
            None if j["last_seen_ts"] is None else Timestamp(j["last_seen_ts"]),
            None
            if j["first_consecutive_seen_ts"] is None
            else Timestamp(j["first_consecutive_seen_ts"]),
            j["recent_lost_answers"],
            j["failed_to_send"],
        )


class LatencyStats:
    fastest: TimestampDuration
    average: TimestampDuration
    slowest: TimestampDuration

    def __init__(
        self,
        fastest: TimestampDuration,
        average: TimestampDuration,
        slowest: TimestampDuration,
    ):
        self.fastest = fastest
        self.average = average
        self.slowest = slowest

    @classmethod
    def from_json(cls, j: dict) -> Self:
        """JSON object hook"""
        return cls(
            TimestampDuration(j["fastest"]),
            TimestampDuration(j["average"]),
            TimestampDuration(j["slowest"]),
        )


class TransferStats:
    total: ByteCount
    maximum: ByteCount
    average: ByteCount
    minimum: ByteCount

    def __init__(
        self,
        total: ByteCount,
        maximum: ByteCount,
        average: ByteCount,
        minimum: ByteCount,
    ):
        self.total = total
        self.maximum = maximum
        self.average = average
        self.minimum = minimum

    @classmethod
    def from_json(cls, j: dict) -> Self:
        """JSON object hook"""
        return cls(
            ByteCount(j["total"]),
            ByteCount(j["maximum"]),
            ByteCount(j["average"]),
            ByteCount(j["minimum"]),
        )


class TransferStatsDownUp:
    down: TransferStats
    up: TransferStats

    def __init__(self, down: TransferStats, up: TransferStats):
        self.down = down
        self.up = up

    @classmethod
    def from_json(cls, j: dict) -> Self:
        """JSON object hook"""
        return cls(TransferStats.from_json(j["down"]), TransferStats.from_json(j["up"]))


class PeerStats:
    time_added: Timestamp
    rpc_stats: RPCStats
    latency: Optional[LatencyStats]
    transfer: TransferStatsDownUp

    def __init__(
        self,
        time_added: Timestamp,
        rpc_stats: RPCStats,
        latency: Optional[LatencyStats],
        transfer: TransferStatsDownUp,
    ):
        self.time_added = time_added
        self.rpc_stats = rpc_stats
        self.latency = latency
        self.transfer = transfer

    @classmethod
    def from_json(cls, j: dict) -> Self:
        """JSON object hook"""
        return cls(
            j["time_added"],
            RPCStats.from_json(j["rpc_stats"]),
            None if j["latency"] is None else LatencyStats.from_json(j["latency"]),
            TransferStatsDownUp.from_json(j["transfer"]),
        )


class PeerTableData:
    node_ids: list[str]
    peer_address: str
    peer_stats: PeerStats

    def __init__(self, node_ids: list[str], peer_address: str, peer_stats: PeerStats):
        self.node_ids = node_ids
        self.peer_address = peer_address
        self.peer_stats = peer_stats

    @classmethod
    def from_json(cls, j: dict) -> Self:
        """JSON object hook"""
        return cls(j["node_ids"], j["peer_address"], PeerStats.from_json(j["peer_stats"]))


class VeilidStateNetwork:
    started: bool
    bps_down: ByteCount
    bps_up: ByteCount
    peers: list[PeerTableData]

    def __init__(
        self,
        started: bool,
        bps_down: ByteCount,
        bps_up: ByteCount,
        peers: list[PeerTableData],
    ):
        self.started = started
        self.bps_down = bps_down
        self.bps_up = bps_up
        self.peers = peers

    @classmethod
    def from_json(cls, j: dict) -> Self:
        """JSON object hook"""
        return cls(
            j["started"],
            ByteCount(j["bps_down"]),
            ByteCount(j["bps_up"]),
            [PeerTableData.from_json(peer) for peer in j["peers"]],
        )


class VeilidStateConfig:
    config: VeilidConfig

    def __init__(self, config: VeilidConfig):
        self.config = config

    @classmethod
    def from_json(cls, j: dict) -> Self:
        """JSON object hook"""
        return cls(VeilidConfig.from_json(j["config"]))


class VeilidState:
    attachment: VeilidStateAttachment
    network: VeilidStateNetwork
    config: VeilidStateConfig

    def __init__(
        self,
        attachment: VeilidStateAttachment,
        network: VeilidStateNetwork,
        config: VeilidStateConfig,
    ):
        self.attachment = attachment
        self.network = network
        self.config = config

    @classmethod
    def from_json(cls, j: dict) -> Self:
        """JSON object hook"""
        return cls(
            VeilidStateAttachment.from_json(j["attachment"]),
            VeilidStateNetwork.from_json(j["network"]),
            VeilidStateConfig.from_json(j["config"]),
        )


class VeilidLog:
    log_level: VeilidLogLevel
    message: str
    backtrace: Optional[str]

    def __init__(self, log_level: VeilidLogLevel, message: str, backtrace: Optional[str]):
        self.log_level = log_level
        self.message = message
        self.backtrace = backtrace

    @classmethod
    def from_json(cls, j: dict) -> Self:
        """JSON object hook"""
        return cls(VeilidLogLevel(j["log_level"]), j["message"], j["backtrace"])


class VeilidAppMessage:
    sender: Optional[TypedKey]
    message: bytes

    def __init__(self, sender: Optional[TypedKey], message: bytes):
        self.sender = sender
        self.message = message

    @classmethod
    def from_json(cls, j: dict) -> Self:
        """JSON object hook"""
        return cls(
            None if j["sender"] is None else TypedKey(j["sender"]),
            urlsafe_b64decode_no_pad(j["message"]),
        )


class VeilidAppCall:
    sender: Optional[TypedKey]
    message: bytes
    call_id: OperationId

    def __init__(self, sender: Optional[TypedKey], message: bytes, call_id: OperationId):
        self.sender = sender
        self.message = message
        self.call_id = call_id

    @classmethod
    def from_json(cls, j: dict) -> Self:
        """JSON object hook"""
        return cls(
            None if j["sender"] is None else TypedKey(j["sender"]),
            urlsafe_b64decode_no_pad(j["message"]),
            OperationId(j["call_id"]),
        )


class VeilidRouteChange:
    dead_routes: list[RouteId]
    dead_remote_routes: list[RouteId]

    def __init__(self, dead_routes: list[RouteId], dead_remote_routes: list[RouteId]):
        self.dead_routes = dead_routes
        self.dead_remote_routes = dead_remote_routes

    @classmethod
    def from_json(cls, j: dict) -> Self:
        """JSON object hook"""
        return cls(
            [RouteId(route) for route in j["dead_routes"]],
            [RouteId(route) for route in j["dead_remote_routes"]],
        )


class VeilidValueChange:
    key: TypedKey
    subkeys: list[ValueSubkey]
    count: int
    value: ValueData

    def __init__(self, key: TypedKey, subkeys: list[ValueSubkey], count: int, value: ValueData):
        self.key = key
        self.subkeys = subkeys
        self.count = count
        self.value = value

    @classmethod
    def from_json(cls, j: dict) -> Self:
        """JSON object hook"""
        return cls(
            TypedKey(j["key"]),
            [ValueSubkey(key) for key in j["subkeys"]],
            j["count"],
            ValueData.from_json(j["value"]),
        )


class VeilidUpdateKind(StrEnum):
    LOG = "Log"
    APP_MESSAGE = "AppMessage"
    APP_CALL = "AppCall"
    ATTACHMENT = "Attachment"
    NETWORK = "Network"
    CONFIG = "Config"
    ROUTE_CHANGE = "RouteChange"
    VALUE_CHANGE = "ValueChange"
    SHUTDOWN = "Shutdown"


VeilidUpdateDetailType = Optional[
    VeilidLog
    | VeilidAppMessage
    | VeilidAppCall
    | VeilidStateAttachment
    | VeilidStateNetwork
    | VeilidStateConfig
    | VeilidRouteChange
    | VeilidValueChange
]


class VeilidUpdate:
    kind: VeilidUpdateKind
    detail: VeilidUpdateDetailType

    def __init__(
        self,
        kind: VeilidUpdateKind,
        detail: VeilidUpdateDetailType,
    ):
        self.kind = kind
        self.detail = detail

    @classmethod
    def from_json(cls, j: dict) -> Self:
        """JSON object hook"""
        kind = VeilidUpdateKind(j["kind"])
        detail: VeilidUpdateDetailType = None
        match kind:
            case VeilidUpdateKind.LOG:
                detail = VeilidLog.from_json(j)
            case VeilidUpdateKind.APP_MESSAGE:
                detail = VeilidAppMessage.from_json(j)
            case VeilidUpdateKind.APP_CALL:
                detail = VeilidAppCall.from_json(j)
            case VeilidUpdateKind.ATTACHMENT:
                detail = VeilidStateAttachment.from_json(j)
            case VeilidUpdateKind.NETWORK:
                detail = VeilidStateNetwork.from_json(j)
            case VeilidUpdateKind.CONFIG:
                detail = VeilidStateConfig.from_json(j)
            case VeilidUpdateKind.ROUTE_CHANGE:
                detail = VeilidRouteChange.from_json(j)
            case VeilidUpdateKind.VALUE_CHANGE:
                detail = VeilidValueChange.from_json(j)
            case VeilidUpdateKind.SHUTDOWN:
                detail = None
            case _:
                raise ValueError("Unknown VeilidUpdateKind")
        return cls(kind, detail)



from dataclasses import dataclass, fields
from enum import StrEnum
from typing import Optional, Self

from .types import TypedKey, TypedSecret, Capability


class VeilidConfigLogLevel(StrEnum):
    OFF = "Off"
    ERROR = "Error"
    WARN = "Warn"
    INFO = "Info"
    DEBUG = "Debug"
    TRACE = "Trace"


@dataclass
class ConfigBase:
    @classmethod
    def from_json(cls, json_data: dict) -> Self:
        """Return an instance of this type from the input data."""
        args = {}
        for field in fields(cls):
            key = field.name
            value = json_data[key]
            try:
                # See if this field's type knows how to load itself from JSON input.
                loader = field.type.from_json
            except AttributeError:
                # No, it doesn't. Use the raw value.
                args[key] = value
            else:
                # Yes, it does. Use the loading function's output.
                args[key] = loader(value)

        return cls(**args)

    def to_json(self) -> dict:
        return self.__dict__


@dataclass
class VeilidConfigCapabilities(ConfigBase):
    disable: list[Capability]


@dataclass
class VeilidConfigProtectedStore(ConfigBase):
    allow_insecure_fallback: bool
    always_use_insecure_storage: bool
    directory: str
    delete: bool
    device_encryption_key_password: str
    new_device_encryption_key_password: Optional[str]


@dataclass
class VeilidConfigTableStore(ConfigBase):
    directory: str
    delete: bool


@dataclass
class VeilidConfigBlockStore(ConfigBase):
    directory: str
    delete: bool


@dataclass
class VeilidConfigRoutingTable(ConfigBase):
    node_id: list[TypedKey]
    node_id_secret: list[TypedSecret]
    bootstrap: list[str]
    limit_over_attached: int
    limit_fully_attached: int
    limit_attached_strong: int
    limit_attached_good: int
    limit_attached_weak: int


@dataclass
class VeilidConfigRPC(ConfigBase):
    concurrency: int
    queue_size: int
    max_timestamp_behind_ms: Optional[int]
    max_timestamp_ahead_ms: Optional[int]
    timeout_ms: int
    max_route_hop_count: int
    default_route_hop_count: int


@dataclass
class VeilidConfigDHT(ConfigBase):
    max_find_node_count: int
    resolve_node_timeout_ms: int
    resolve_node_count: int
    resolve_node_fanout: int
    get_value_timeout_ms: int
    get_value_count: int
    get_value_fanout: int
    set_value_timeout_ms: int
    set_value_count: int
    set_value_fanout: int
    min_peer_count: int
    min_peer_refresh_time_ms: int
    validate_dial_info_receipt_time_ms: int
    local_subkey_cache_size: int
    local_max_subkey_cache_memory_mb: int
    remote_subkey_cache_size: int
    remote_max_records: int
    remote_max_subkey_cache_memory_mb: int
    remote_max_storage_space_mb: int


@dataclass
class VeilidConfigTLS(ConfigBase):
    certificate_path: str
    private_key_path: str
    connection_initial_timeout_ms: int


@dataclass
class VeilidConfigHTTPS(ConfigBase):
    enabled: bool
    listen_address: str
    path: str
    url: Optional[str]


@dataclass
class VeilidConfigHTTP(ConfigBase):
    enabled: bool
    listen_address: str
    path: str
    url: Optional[str]


@dataclass
class VeilidConfigApplication(ConfigBase):
    https: VeilidConfigHTTPS
    http: VeilidConfigHTTP


@dataclass
class VeilidConfigUDP(ConfigBase):
    enabled: bool
    socket_pool_size: int
    listen_address: str
    public_address: Optional[str]


@dataclass
class VeilidConfigTCP(ConfigBase):
    connect: bool
    listen: bool
    max_connections: int
    listen_address: str
    public_address: Optional[str]


@dataclass
class VeilidConfigWS(ConfigBase):
    connect: bool
    listen: bool
    max_connections: int
    listen_address: str
    path: str
    url: Optional[str]


@dataclass
class VeilidConfigWSS(ConfigBase):
    connect: bool
    listen: bool
    max_connections: int
    listen_address: str
    path: str
    url: Optional[str]


@dataclass
class VeilidConfigProtocol(ConfigBase):
    udp: VeilidConfigUDP
    tcp: VeilidConfigTCP
    ws: VeilidConfigWS
    wss: VeilidConfigWSS


@dataclass
class VeilidConfigNetwork(ConfigBase):
    connection_initial_timeout_ms: int
    connection_inactivity_timeout_ms: int
    max_connections_per_ip4: int
    max_connections_per_ip6_prefix: int
    max_connections_per_ip6_prefix_size: int
    max_connection_frequency_per_min: int
    client_whitelist_timeout_ms: int
    reverse_connection_receipt_time_ms: int
    hole_punch_receipt_time_ms: int
    network_key_password: Optional[str]
    routing_table: VeilidConfigRoutingTable
    rpc: VeilidConfigRPC
    dht: VeilidConfigDHT
    upnp: bool
    detect_address_changes: bool
    restricted_nat_retries: int
    tls: VeilidConfigTLS
    application: VeilidConfigApplication
    protocol: VeilidConfigProtocol


@dataclass
class VeilidConfig(ConfigBase):
    program_name: str
    namespace: str
    capabilities: VeilidConfigCapabilities
    protected_store: VeilidConfigProtectedStore
    table_store: VeilidConfigTableStore
    block_store: VeilidConfigBlockStore
    network: VeilidConfigNetwork



from enum import StrEnum
from typing import Self


class Operation(StrEnum):
    CONTROL = "Control"
    GET_STATE = "GetState"
    ATTACH = "Attach"
    DETACH = "Detach"
    NEW_PRIVATE_ROUTE = "NewPrivateRoute"
    NEW_CUSTOM_PRIVATE_ROUTE = "NewCustomPrivateRoute"
    IMPORT_REMOTE_PRIVATE_ROUTE = "ImportRemotePrivateRoute"
    RELEASE_PRIVATE_ROUTE = "ReleasePrivateRoute"
    APP_CALL_REPLY = "AppCallReply"
    NEW_ROUTING_CONTEXT = "NewRoutingContext"
    ROUTING_CONTEXT = "RoutingContext"
    OPEN_TABLE_DB = "OpenTableDb"
    DELETE_TABLE_DB = "DeleteTableDb"
    TABLE_DB = "TableDb"
    TABLE_DB_TRANSACTION = "TableDbTransaction"
    GET_CRYPTO_SYSTEM = "GetCryptoSystem"
    BEST_CRYPTO_SYSTEM = "BestCryptoSystem"
    CRYPTO_SYSTEM = "CryptoSystem"
    VERIFY_SIGNATURES = "VerifySignatures"
    GENERATE_SIGNATURES = "GenerateSignatures"
    GENERATE_KEY_PAIR = "GenerateKeyPair"
    NOW = "Now"
    DEBUG = "Debug"
    VEILID_VERSION_STRING = "VeilidVersionString"
    VEILID_VERSION = "VeilidVersion"


class RoutingContextOperation(StrEnum):
    INVALID_ID = "InvalidId"
    RELEASE = "Release"
    WITH_PRIVACY = "WithPrivacy"
    WITH_CUSTOM_PRIVACY = "WithCustomPrivacy"
    WITH_SEQUENCING = "WithSequencing"
    APP_CALL = "AppCall"
    APP_MESSAGE = "AppMessage"
    CREATE_DHT_RECORD = "CreateDhtRecord"
    OPEN_DHT_RECORD = "OpenDhtRecord"
    CLOSE_DHT_RECORD = "CloseDhtRecord"
    DELETE_DHT_RECORD = "DeleteDhtRecord"
    GET_DHT_VALUE = "GetDhtValue"
    SET_DHT_VALUE = "SetDhtValue"
    WATCH_DHT_VALUES = "WatchDhtValues"
    CANCEL_DHT_WATCH = "CancelDhtWatch"


class TableDbOperation(StrEnum):
    INVALID_ID = "InvalidId"
    RELEASE = "Release"
    GET_COLUMN_COUNT = "GetColumnCount"
    GET_KEYS = "GetKeys"
    TRANSACT = "Transact"
    STORE = "Store"
    LOAD = "Load"
    DELETE = "Delete"


class TableDbTransactionOperation(StrEnum):
    INVALID_ID = "InvalidId"
    COMMIT = "Commit"
    ROLLBACK = "Rollback"
    STORE = "Store"
    DELETE = "Delete"


class CryptoSystemOperation(StrEnum):
    INVALID_ID = "InvalidId"
    RELEASE = "Release"
    CACHED_DH = "CachedDh"
    COMPUTE_DH = "ComputeDh"
    RANDOM_BYTES = "RandomBytes"
    DEFAULT_SALT_LENGTH = "DefaultSaltLength"
    HASH_PASSWORD = "HashPassword"
    VERIFY_PASSWORD = "VerifyPassword"
    DERIVE_SHARED_SECRET = "DeriveSharedSecret"
    RANDOM_NONCE = "RandomNonce"
    RANDOM_SHARED_SECRET = "RandomSharedSecret"
    GENERATE_KEY_PAIR = "GenerateKeyPair"
    GENERATE_HASH = "GenerateHash"
    VALIDATE_KEY_PAIR = "ValidateKeyPair"
    VALIDATE_HASH = "ValidateHash"
    DISTANCE = "Distance"
    SIGN = "Sign"
    VERIFY = "Verify"
    AEAD_OVERHEAD = "AeadOverhead"
    DECRYPT_AEAD = "DecryptAead"
    ENCRYPT_AEAD = "EncryptAead"
    CRYPT_NO_AUTH = "CryptNoAuth"


class RecvMessageType(StrEnum):
    RESPONSE = "Response"
    UPDATE = "Update"



import base64
import json
from enum import StrEnum
from functools import total_ordering
from typing import Any, Optional, Self, Tuple

####################################################################


def urlsafe_b64encode_no_pad(b: bytes) -> str:
    """
    Removes any `=` used as padding from the encoded string.
    """
    return base64.urlsafe_b64encode(b).decode().rstrip("=")


def urlsafe_b64decode_no_pad(s: str) -> bytes:
    """
    Adds back in the required padding before decoding.
    """
    padding = 4 - (len(s) % 4)
    s = s + ("=" * padding)
    return base64.urlsafe_b64decode(s)


class VeilidJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, bytes):
            return urlsafe_b64encode_no_pad(o)
        if hasattr(o, "to_json") and callable(o.to_json):
            return o.to_json()
        return json.JSONEncoder.default(self, o)

    @staticmethod
    def dumps(req: Any, *args, **kwargs) -> str:
        return json.dumps(req, cls=VeilidJSONEncoder, *args, **kwargs)


####################################################################


class VeilidLogLevel(StrEnum):
    ERROR = "Error"
    WARN = "Warn"
    INFO = "Info"
    DEBUG = "Debug"
    TRACE = "Trace"


class CryptoKind(StrEnum):
    CRYPTO_KIND_NONE = "NONE"
    CRYPTO_KIND_VLD0 = "VLD0"


class Capability(StrEnum):
    CAP_ROUTE = "ROUT"
    CAP_TUNNEL = "TUNL"
    CAP_SIGNAL = "SGNL"
    CAP_RELAY = "RLAY"
    CAP_VALIDATE_DIAL_INFO = "DIAL"
    CAP_DHT = "DHTV"
    CAP_APPMESSAGE = "APPM"
    CAP_BLOCKSTORE = "BLOC"


class Stability(StrEnum):
    LOW_LATENCY = "LowLatency"
    RELIABLE = "Reliable"


class Sequencing(StrEnum):
    NO_PREFERENCE = "NoPreference"
    PREFER_ORDERED = "PreferOrdered"
    ENSURE_ORDERED = "EnsureOrdered"


class DHTSchemaKind(StrEnum):
    DFLT = "DFLT"
    SMPL = "SMPL"


class SafetySelectionKind(StrEnum):
    UNSAFE = "Unsafe"
    SAFE = "Safe"


####################################################################


class Timestamp(int):
    pass


class TimestampDuration(int):
    pass


class ByteCount(int):
    pass


class OperationId(str):
    pass


class RouteId(str):
    pass


class EncodedString(str):
    def to_bytes(self) -> bytes:
        return urlsafe_b64decode_no_pad(self)

    @classmethod
    def from_bytes(cls, b: bytes) -> Self:
        return cls(urlsafe_b64encode_no_pad(b))


class CryptoKey(EncodedString):
    pass


class CryptoKeyDistance(CryptoKey):
    pass


class PublicKey(CryptoKey):
    pass


class SecretKey(CryptoKey):
    pass


class SharedSecret(CryptoKey):
    pass


class HashDigest(CryptoKey):
    pass


class Signature(EncodedString):
    pass


class Nonce(EncodedString):
    pass


class KeyPair(str):
    @classmethod
    def from_parts(cls, key: PublicKey, secret: SecretKey) -> Self:
        return cls(f"{key}:{secret}")

    def key(self) -> PublicKey:
        return PublicKey(self.split(":", 1)[0])

    def secret(self) -> SecretKey:
        return SecretKey(self.split(":", 1)[1])

    def to_parts(self) -> Tuple[PublicKey, SecretKey]:
        public, secret = self.split(":", 1)
        return (PublicKey(public), SecretKey(secret))


class CryptoTyped(str):
    def kind(self) -> CryptoKind:
        if self[4] != ":":
            raise ValueError("Not CryptoTyped")
        return CryptoKind(self[0:4])

    def _value(self) -> str:
        if self[4] != ":":
            raise ValueError("Not CryptoTyped")
        return self[5:]


class TypedKey(CryptoTyped):
    @classmethod
    def from_value(cls, kind: CryptoKind, value: PublicKey) -> Self:
        return cls(f"{kind}:{value}")

    def value(self) -> PublicKey:
        return PublicKey(self._value())


class TypedSecret(CryptoTyped):
    @classmethod
    def from_value(cls, kind: CryptoKind, value: SecretKey) -> Self:
        return cls(f"{kind}:{value}")

    def value(self) -> SecretKey:
        return SecretKey(self._value())


class TypedKeyPair(CryptoTyped):
    @classmethod
    def from_value(cls, kind: CryptoKind, value: KeyPair) -> Self:
        return cls(f"{kind}:{value}")

    def value(self) -> KeyPair:
        return KeyPair(self._value())


class TypedSignature(CryptoTyped):
    @classmethod
    def from_value(cls, kind: CryptoKind, value: Signature) -> Self:
        return cls(f"{kind}:{value}")

    def value(self) -> Signature:
        return Signature(self._value())


class ValueSubkey(int):
    pass


class ValueSeqNum(int):
    pass


####################################################################


@total_ordering
class VeilidVersion:
    _major: int
    _minor: int
    _patch: int

    def __init__(self, major: int, minor: int, patch: int):
        self._major = major
        self._minor = minor
        self._patch = patch

    def __lt__(self, other):
        if other is None:
            return False
        if self._major < other._major:
            return True
        if self._major > other._major:
            return False
        if self._minor < other._minor:
            return True
        if self._minor > other._minor:
            return False
        if self._patch < other._patch:
            return True
        return False

    def __eq__(self, other):
        return (
            isinstance(other, VeilidVersion)
            and self.data == other.data
            and self.seq == other.seq
            and self.writer == other.writer
        )

    @property
    def major(self):
        return self._major

    @property
    def minor(self):
        return self._minor

    @property
    def patch(self):
        return self._patch


class NewPrivateRouteResult:
    route_id: RouteId
    blob: bytes

    def __init__(self, route_id: RouteId, blob: bytes):
        self.route_id = route_id
        self.blob = blob

    def to_tuple(self) -> Tuple[RouteId, bytes]:
        return (self.route_id, self.blob)

    @classmethod
    def from_json(cls, j: dict) -> Self:
        return cls(RouteId(j["route_id"]), urlsafe_b64decode_no_pad(j["blob"]))


class DHTSchemaSMPLMember:
    m_key: PublicKey
    m_cnt: int

    def __init__(self, m_key: PublicKey, m_cnt: int):
        self.m_key = m_key
        self.m_cnt = m_cnt

    @classmethod
    def from_json(cls, j: dict) -> Self:
        return cls(PublicKey(j["m_key"]), j["m_cnt"])

    def to_json(self) -> dict:
        return self.__dict__


class DHTSchema:
    kind: DHTSchemaKind

    def __init__(self, kind: DHTSchemaKind, **kwargs):
        self.kind = kind
        for k, v in kwargs.items():
            setattr(self, k, v)

    @classmethod
    def dflt(cls, o_cnt: int) -> Self:
        return cls(DHTSchemaKind.DFLT, o_cnt=o_cnt)

    @classmethod
    def smpl(cls, o_cnt: int, members: list[DHTSchemaSMPLMember]) -> Self:
        return cls(DHTSchemaKind.SMPL, o_cnt=o_cnt, members=members)

    @classmethod
    def from_json(cls, j: dict) -> Self:
        if DHTSchemaKind(j["kind"]) == DHTSchemaKind.DFLT:
            return cls.dflt(j["o_cnt"])
        if DHTSchemaKind(j["kind"]) == DHTSchemaKind.SMPL:
            return cls.smpl(
                j["o_cnt"],
                [DHTSchemaSMPLMember.from_json(member) for member in j["members"]],
            )
        raise Exception("Unknown DHTSchema kind", j["kind"])

    def to_json(self) -> dict:
        return self.__dict__


class DHTRecordDescriptor:
    key: TypedKey
    owner: PublicKey
    owner_secret: Optional[SecretKey]
    schema: DHTSchema

    def __init__(
        self,
        key: TypedKey,
        owner: PublicKey,
        owner_secret: Optional[SecretKey],
        schema: DHTSchema,
    ):
        self.key = key
        self.owner = owner
        self.owner_secret = owner_secret
        self.schema = schema

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}(key={self.key!r})>"

    @classmethod
    def from_json(cls, j: dict) -> Self:
        return cls(
            TypedKey(j["key"]),
            PublicKey(j["owner"]),
            None if j["owner_secret"] is None else SecretKey(j["owner_secret"]),
            DHTSchema.from_json(j["schema"]),
        )

    def to_json(self) -> dict:
        return self.__dict__


@total_ordering
class ValueData:
    seq: ValueSeqNum
    data: bytes
    writer: PublicKey

    def __init__(self, seq: ValueSeqNum, data: bytes, writer: PublicKey):
        self.seq = seq
        self.data = data
        self.writer = writer

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}(seq={self.seq!r}, data={self.data!r}, writer={self.writer!r})>"

    def __lt__(self, other):
        if other is None:
            return True
        if self.data < other.data:
            return True
        if self.data > other.data:
            return False
        if self.seq < other.seq:
            return True
        if self.seq > other.seq:
            return False
        if self.writer < other.writer:
            return True
        return False

    def __eq__(self, other):
        return (
            isinstance(other, ValueData)
            and self.data == other.data
            and self.seq == other.seq
            and self.writer == other.writer
        )

    @classmethod
    def from_json(cls, j: dict) -> Self:
        return cls(
            ValueSeqNum(j["seq"]),
            urlsafe_b64decode_no_pad(j["data"]),
            PublicKey(j["writer"]),
        )

    def to_json(self) -> dict:
        return self.__dict__


####################################################################


class SafetySpec:
    preferred_route: Optional[RouteId]
    hop_count: int
    stability: Stability
    sequencing: Sequencing

    def __init__(
        self,
        preferred_route: Optional[RouteId],
        hop_count: int,
        stability: Stability,
        sequencing: Sequencing,
    ):
        self.preferred_route = preferred_route
        self.hop_count = hop_count
        self.stability = stability
        self.sequencing = sequencing

    @classmethod
    def from_json(cls, j: dict) -> Self:
        return cls(
            RouteId(j["preferred_route"]) if "preferred_route" in j else None,
            j["hop_count"],
            Stability(j["stability"]),
            Sequencing(j["sequencing"]),
        )

    def to_json(self) -> dict:
        return self.__dict__


class SafetySelection:
    kind: SafetySelectionKind

    def __init__(self, kind: SafetySelectionKind, **kwargs):
        self.kind = kind
        for k, v in kwargs.items():
            setattr(self, k, v)

    @classmethod
    def unsafe(cls, sequencing: Sequencing) -> Self:
        return cls(SafetySelectionKind.UNSAFE, sequencing=sequencing)

    @classmethod
    def safe(cls, safety_spec: SafetySpec) -> Self:
        return cls(SafetySelectionKind.SAFE, safety_spec=safety_spec)

    @classmethod
    def from_json(cls, j: dict) -> Self:
        if "Safe" in j:
            return cls.safe(SafetySpec.from_json(j["Safe"]))
        elif "Unsafe" in j:
            return cls.unsafe(Sequencing(j["Unsafe"]))
        raise Exception("Invalid SafetySelection")

    def to_json(self) -> dict:
        if self.kind == SafetySelectionKind.UNSAFE:
            return {"Unsafe": self.sequencing}
        elif self.kind == SafetySelectionKind.SAFE:
            return {"Safe": self.safety_spec.to_json()}
        else:
            raise Exception("Invalid SafetySelection")
